import os
import pathlib
import numpy as np

def dict_slice(d, idx):
    """
    Slice a dictionary of jagged arrays, returning a dictionary of singleton numpy arrays.
    """
    return {k: np.array(v[idx]).reshape(1, -1) for k, v in d.items()}

def get_cpu_info():
    # figure out which model to load. check if current cpu is avx2, avx512, avx512_vnni, or arm64
    try:
        cpu_info = print(pathlib.Path("/proc/cpuinfo").read_text())
        if "avx512" in cpu_info and "vnni" in cpu_info:
            return "avx512_vnni"
        elif "avx512" in cpu_info:
            return "avx512"
        elif "avx2" in cpu_info or "x86-64" in cpu_info:
            return "avx2"
        elif "aarch64" in cpu_info.lower():
            return "arm64"
        else:
            return None
    except Exception as e:
        print(f"Failed to read /proc/cpuinfo: {e}")
        try:
            # maybe we're on a mac, use sysctl
            cpu_info = os.popen("sysctl -a").read()
            if "intel" in cpu_info.lower():
                if "avx512" in cpu_info.lower() and "vnni" in cpu_info.lower():
                    return "avx512_vnni"
                elif "avx512" in cpu_info.lower():
                    return "avx512"
                elif "avx2" in cpu_info.lower():
                    return "avx2"
                else:
                    return None
            elif (
                "apple m1" in cpu_info.lower()
                or "apple m2" in cpu_info.lower()
                or "apple m3" in cpu_info.lower()
            ):
                return "arm64"
            else:
                return None
        except Exception as e:
            print(f"Failed to run sysctl -a: {e}")
            return None


def find_onnx_model(repo_path: str):
    cpu_type = get_cpu_info()
    
    # check if there's an 'onnx' folder in the repo
    if os.path.exists(os.path.join(repo_path, "onnx")):
        onnx_models = os.listdir(os.path.join(repo_path, "onnx"))
        onnx_models = [os.path.join(repo_path, "onnx", model) for model in onnx_models if model.endswith(".onnx")]

    # otherwise check for onnx models in the root of the repo
    else:
        onnx_models = os.listdir(repo_path)
        onnx_models = [os.path.join(repo_path, model) for model in onnx_models if model.endswith(".onnx")]

    # try to select the model based on the cpu type
    matching_cpu_models = [model for model in onnx_models if cpu_type is not None and cpu_type in model]
    if len(matching_cpu_models) > 0:
        return matching_cpu_models[0]
    
    # if none matching the cpu type, prefer a quantized model
    quantized_models = [model for model in onnx_models if "quantized" in model]
    if len(quantized_models) > 0:
        return quantized_models[0]
    
    # if no quantized model, just return the first model
    if len(onnx_models) > 0:
        return onnx_models[0]
    else:
        return None


import os
import json
import abc
import tqdm
import shutil
import tempfile
from typing import Literal, Optional, Union

import numpy as np
from huggingface_hub import snapshot_download
from scipy.sparse import csr_matrix
from .registry import registry
from .utils import dict_slice, find_onnx_model

class EmbeddingModelBase(abc.ABC):
    def __init__(
        self,
        onnx_path: str,
        tokenizer_path: str,
        max_length: int,
        pooling_strategy: Literal["mean", "first", "cls", "max", None],
        normalize: bool,
        intra_op_num_threads: int = 0,
        thread_spinning: bool = True,
    ):
        """
        This assumes the model file is already downloaded.
        Use classmethods / utilities to load from registry or Hub.
        Recommended to also pre-download tokenizer to avoid cold start.
        """
        import onnxruntime as ort
        from transformers import AutoTokenizer
        sess_options = ort.SessionOptions()
        # options to make it faster
        sess_options.intra_op_num_threads = intra_op_num_threads
        sess_options.add_session_config_entry(
            "session.intra_op.allow_spinning",
            "1" if thread_spinning else "0",
        )

        self.max_length = max_length
        self.pooling_strategy = pooling_strategy
        self.normalize = normalize
        self.session = ort.InferenceSession(onnx_path)
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)
    
    @classmethod
    def from_registry(
        cls,
        model_id: str,
        normalize: bool = True,
        destination: Optional[str] = None,
        intra_op_num_threads: int = 0,
        thread_spinning: bool = True,
    ):
        """
        Downloads a model from the pre-selected registry and returns an instance.
        """
        if destination is not None:
            os.makedirs(destination, exist_ok=True)
            if not os.path.exists(os.path.join(destination, "model.onnx")):
                cls.download_from_registry(model_id, destination)
                instance = cls(
                    os.path.join(destination, "model.onnx"),
                    destination,
                    max_length=registry[model_id]["max_length"],
                    pooling_strategy=registry[model_id]["pooling_strategy"],
                    normalize=normalize,
                    intra_op_num_threads=intra_op_num_threads,
                    thread_spinning=thread_spinning,
                )
        else:
            with tempfile.TemporaryDirectory() as tmpdir:
                cls.download_from_registry(model_id, tmpdir)
                instance = cls(
                    os.path.join(tmpdir, "model.onnx"),
                    tmpdir,
                    max_length=registry[model_id]["max_length"],
                    pooling_strategy=registry[model_id]["pooling_strategy"],
                    normalize=normalize,
                    intra_op_num_threads=intra_op_num_threads,
                    thread_spinning=thread_spinning,
                )
        return instance

    @classmethod
    def from_pretrained(cls, model_name_or_path: str):
        if os.path.exists(model_name_or_path):
            model_path = model_name_or_path
        else:
            model_path = snapshot_download(model_name_or_path, allow_patterns=["*.onnx", "*.json", "*.txt"])
        with open(os.path.join(model_path, "config.json")) as f:
            config = json.load(f)
        onnx_model_path = find_onnx_model(model_path)
        if onnx_model_path is None:
            raise ValueError(f"Could not find an ONNX model in repo {model_name_or_path}")
        
        # try to load the sentence transformers pooling config
        try:
            st_config = json.load(open(os.path.join(model_path, "1_Pooling/config.json")))
            if st_config["pooling_mode_cls_token"]:
                pooling_strategy = "first" # for mlx_embedding_models, cls refers to pooler output
            elif st_config["pooling_mode_mean_tokens"]:
                pooling_strategy = "mean"
            elif st_config["pooling_mode_max_tokens"]:
                pooling_strategy = "max"
            else:
                raise ValueError("Unsupported or missing pooling strategy.")
        except FileNotFoundError:
            print("WARNING: No sentence-transformers config found. Using CLS token for pooling.")
            pooling_strategy = "first"

        return cls(
            onnx_model_path,
            model_path,
            max_length=config.get("max_position_embeddings", 512),
            pooling_strategy=pooling_strategy,
            normalize=True,
        )



    @staticmethod
    def download_from_registry(model_id: str, destination: str):
        """
        Downloads a model from the pre-selected registry.
        """
        import huggingface_hub
        from transformers import AutoTokenizer
        os.makedirs(destination, exist_ok=True)
        with tempfile.TemporaryDirectory() as tmpdir:
            onnx_file = huggingface_hub.hf_hub_download(
                repo_id=registry[model_id]["repo"],
                filename=registry[model_id]["path_in_repo"],
                local_dir=tmpdir,
                local_dir_use_symlinks=False,  # we want the actual onnx files there
            )
            shutil.move(onnx_file, os.path.join(destination, "model.onnx"))
        tokenizer = AutoTokenizer.from_pretrained(registry[model_id]["repo"])
        tokenizer.save_pretrained(destination)

    @staticmethod
    def download_from_hub(
        model_repo: str, model_path_in_repo: str, tokenizer_repo: str, destination: str
    ):
        """
        Downloads a model from a custom Hub repo, for those not in the registry.
        """
        import huggingface_hub
        from transformers import AutoTokenizer

        with tempfile.TemporaryDirectory() as tmpdir:
            onnx_file = huggingface_hub.hf_hub_download(
                repo_id=model_repo,
                filename=model_path_in_repo,
                local_dir=tmpdir,
                local_dir_use_symlinks=False,  # we want the actual onnx files there
            )
            shutil.move(onnx_file, os.path.join(destination, "model.onnx"))
        tokenizer = AutoTokenizer.from_pretrained(tokenizer_repo)
        tokenizer.save_pretrained(destination)
    
    def _forward_one(
        self,
        inputs: dict[str, np.ndarray]
    ):
        outputs = self.session.run(None, inputs)
        if len(outputs) == 2:
            hidden_states, pooler_output = outputs
        else:
            hidden_states, pooler_output = outputs[0], None
        return hidden_states, pooler_output

    def _forward_batch(
        self,
        inputs: dict[str, list[list[int]]],
        show_progress: bool = True,
        chunk_size: Optional[int] = None,
    ):
        hidden_states_list = []
        pooler_output_list = []
        num_inputs = len(inputs["input_ids"])
        for i in tqdm.tqdm(range(num_inputs), disable=not show_progress):
            if chunk_size is not None:
                input_ids = inputs["input_ids"][i]
                attention_mask = inputs["attention_mask"][i]
                token_type_ids = inputs["token_type_ids"][i] if "token_type_ids" in inputs else None
                input_ids_chunks = [
                    input_ids[i : i + chunk_size] for i in range(0, len(input_ids), chunk_size)
                ]
                attention_mask_chunks = [
                    attention_mask[i : i + chunk_size]
                    for i in range(0, len(attention_mask), chunk_size)
                ]
                if token_type_ids is not None:
                    token_type_ids_chunks = [
                        token_type_ids[i : i + chunk_size]
                        for i in range(0, len(token_type_ids), chunk_size)
                    ]
                else:
                    token_type_ids_chunks = [None] * len(input_ids_chunks)

                hidden_states_chunks = []
                for chunk_ids, chunk_mask, chunk_ttids in zip(input_ids_chunks, attention_mask_chunks, token_type_ids_chunks):
                    chunk_inputs = {
                        "input_ids": np.array([chunk_ids]),
                        "attention_mask": np.array([chunk_mask]),
                    }
                    if chunk_ttids is not None:
                        chunk_inputs["token_type_ids"] = np.array([chunk_ttids])
                    hidden_states_chunk, pooler_output_chunk = self._forward_one(chunk_inputs)
                    hidden_states_chunks.append(hidden_states_chunk) # B, chunk_size, D

                hidden_states = np.concatenate(hidden_states_chunks, axis=1)
                pooler_output = None # can't aggregate pooler_output for chunked
            else:
                hidden_states, pooler_output = self._forward_one(dict_slice(inputs, i))
            
            hidden_states_list.append(hidden_states)
            pooler_output_list.append(pooler_output)

        return hidden_states_list, pooler_output_list

    def _pool(
        self,
        last_hidden_state: np.ndarray,  # B, L, D
        pooler_output: Optional[np.ndarray] = None,  # B, D
        mask: Optional[np.ndarray] = None,  # B, L
    ):
        # hiddens: B, L, D; mask: B, L
        if mask is None:
            mask = np.ones(last_hidden_state.shape[:2])
        if self.pooling_strategy == "mean":
            pooled = np.sum(
                last_hidden_state * np.expand_dims(mask, -1), axis=1
            ) / np.sum(mask, axis=-1, keepdims=True)
        if self.pooling_strategy == "max":
            pooled = np.max(
                last_hidden_state + (1 - np.expand_dims(mask, -1)) * -1e6, axis=1
            )
        elif self.pooling_strategy == "first":
            pooled = last_hidden_state[:, 0, :]
        elif self.pooling_strategy == "cls":
            if pooler_output is None:
                # use first token w/ no pooling linear layer
                pooled = last_hidden_state[:, 0, :]
            else:
                pooled = pooler_output
        else:
            raise NotImplementedError(
                f"pooling strategy {self.pooling_strategy} not implemented"
            )
        if self.normalize:
            pooled = pooled / np.linalg.norm(pooled, axis=-1, keepdims=True)

        return pooled

    def encode(
        self,
        texts: list[str],
        return_numpy=True,
        show_progress=True,
        **kwargs,
    ):
        pass
    
    def __call__(self, texts: list[str], return_numpy=False, show_progress=True, **kwargs):
        if isinstance(texts, str):
            return self.encode([texts], return_numpy=return_numpy, show_progress=show_progress, **kwargs)[0]
        elif isinstance(texts, list):
            return self.encode(texts, return_numpy=return_numpy, show_progress=show_progress, **kwargs)

class EmbeddingModel(EmbeddingModelBase):
    def encode(
        self,
        texts: list[str],
        return_numpy: bool = True,
        show_progress: bool = True,
    ):
        inputs = self.tokenizer(
            texts,
            truncation=True,
            padding=False,
            max_length=self.max_length,
        ) # dont return tensors, this adds unnecessary padding
        hidden_states_list, pooler_output_list = self._forward_batch(inputs, show_progress=show_progress)
        output_embs = [
            self._pool(hidden_states, pooler_output).flatten()
            for hidden_states, pooler_output in zip(hidden_states_list, pooler_output_list)
        ]
        if return_numpy:
            return np.array(output_embs)
        return output_embs

ONNXEmbeddingModel = EmbeddingModel

class SpladeModel(EmbeddingModelBase):

    @staticmethod
    def _create_sparse_embedding(
        activations: np.ndarray,
        max_dims: int,
    ):
        B, V = activations.shape
        topk_indices = np.argsort(activations, axis=-1)[:, -max_dims:] # B, max_dims
        sparse_embeddings = np.zeros((B, V), dtype=np.float32)
        for i in range(B):
            sparse_embeddings[i, topk_indices[i]] = activations[i, topk_indices[i]]

        return sparse_embeddings
    
    def encode(
        self,
        texts: list[str],
        return_numpy: bool = True,
        show_progress: bool = True,
        max_dims: Union[int, str] = "auto",
        return_sparse: bool = False,
        chunk_size: Optional[int] = None,
    ):
        if return_numpy and return_sparse:
            raise ValueError("Can't return both numpy and sparse embeddings")
        inputs = self.tokenizer(
            texts,
            truncation=True,
            padding=False,
            max_length=self.max_length,
        ) # dont return tensors, this adds unnecessary padding
        hidden_states_list, _ = self._forward_batch(inputs, show_progress=show_progress, chunk_size=chunk_size)
        # relu + max pool
        sparse_activations = np.array([
            self._pool(np.maximum(hidden_states, 0)).flatten()
            for hidden_states in hidden_states_list
        ]) # B, V

        # topk
        max_dims = self.max_length if max_dims == "auto" else max_dims
        sparse_embs = self._create_sparse_embedding(sparse_activations, max_dims)

        # Select top-k activations
        if return_numpy:
            return sparse_embs
        elif return_sparse:
            return csr_matrix(sparse_embs)
        return sparse_embs.tolist()


import os
from typing import Literal
from tqdm.auto import trange
from .utils import dict_slice
import numpy as np

class CrossEncoder:
    def __init__(
        self,
        onnx_path: str,
        tokenizer_path: str,
        max_length: int,
        intra_op_num_threads: int = 0,
        thread_spinning: bool = True,
    ):
        """
        This assumes the model file is already downloaded.
        Use classmethods / utilities to load from registry or Hub.
        Recommended to also pre-download tokenizer to avoid cold start.
        """
        import onnxruntime as ort
        from transformers import AutoTokenizer
        sess_options = ort.SessionOptions()
        # options to make it faster
        sess_options.intra_op_num_threads = intra_op_num_threads
        sess_options.add_session_config_entry(
            "session.intra_op.allow_spinning",
            "1" if thread_spinning else "0",
        )

        self.max_length = max_length
        self.session = ort.InferenceSession(onnx_path)
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)
        os.environ["TOKENIZERS_PARALLELISM"] = "false"

    def rerank(self, texts: list[str], candidate_labels: list[list[str]]):
        """
        Rerank the candidate labels for each text.
        """
        results = []
        # allow for candidate labels to be (label, score) tuples - in that case, extract the labels
        if not isinstance(candidate_labels[0][0], str):
            candidate_labels = [[tup[0] for tup in lab_list] for lab_list in candidate_labels]

        # flatten samples while keeping track of the original text index
        samples = []
        for i, (text, labels) in enumerate(zip(texts, candidate_labels)):
            samples.extend([
                (i, text, label) for label in labels
            ])

        # tokenize everything at once
        tokenized = self.tokenizer.batch_encode_plus(
            [(text, label) for _, text, label in samples],
            add_special_tokens=True, 
            max_length=512, 
            # no padding
            padding=False,
            truncation='only_first', # truncate only query, not the label
            return_tensors=None
        )
        # process one at a time (best for cpu)
        all_logits = []
        for i in trange(0, len(samples)):
            batch = dict_slice(tokenized, i)
            logits = self.session.run(None, batch)[0][0] # list of length 2
            all_logits.append(logits)
        
        all_logits = np.array(all_logits)

        # numerically stable softmax
        all_probs = np.exp(all_logits - all_logits.max(axis=1, keepdims=True))
        all_probs /= all_probs.sum(axis=1, keepdims=True)

        # zip with original index and group
        # group by index
        grouped = {}
        assert len(samples) == len(all_probs), "Mismatch in number of samples and output probabilities"
        for orig_sample, probs in zip(samples, all_probs):
            i, text, label = orig_sample
            if i not in grouped:
                grouped[i] = []
            grouped[i].append((label, probs[1]))
        
        results = [
            sorted(grouped[i], key=lambda x: x[1], reverse=True) for i in range(len(texts))
        ]
        
        return results

    # def predict_scores(self, texts: list[str], candidate_labels: Optional[list[list[str]]] = None):
    #     """
    #     Predict the relevance scores for each text and label pair. If provided, will only 
    #     rerank the candidate labels.
    #     """
    #     device = "cuda" if torch.cuda.is_available() else "cpu"
    #     tokenizer = AutoTokenizer.from_pretrained(backbones[self.base_model])
    #     model = self.model
    #     model.eval()
    #     with torch.no_grad():
    #         batch = prepare_batch(texts, candidate_labels, tokenizer, device)
    #         scores = model(**batch).logits
    #     return scores


registry = {
    # 3 layers, 384-dim
    "bge-micro": {
        "repo": "TaylorAI/bge-micro-v2",
        "path_in_repo": "onnx/model_quantized.onnx",
        "max_length": 512,
        "pooling_strategy": "mean",
        "ndim": 384,
    },
    # 6 layers, 384-dim
    "gte-tiny": {
        "repo": "TaylorAI/gte-tiny",
        "path_in_repo": "onnx/model_quantized.onnx",
        "max_length": 512,
        "pooling_strategy": "mean",
        "ndim": 384,
    },
    "minilm-l6": {
        "repo": "Xenova/all-MiniLM-L6-v2",
        "path_in_repo": "onnx/model_quantized.onnx",
        "max_length": 512,
        "pooling_strategy": "mean",
        "ndim": 384,
    },
    "snowflake-xs": {
        "repo": "Snowflake/snowflake-arctic-embed-xs",
        "path_in_repo": "onnx/model_quantized.onnx",
        "max_length": 512,
        "pooling_strategy": "cls",
        "normalize": True,
        "ndim": 384,
    },
    # 12 layers, 384-dim
    "minilm-l12": {
        "repo": "Xenova/all-MiniLM-L12-v2",
        "path_in_repo": "onnx/model_quantized.onnx",
        "max_length": 512,
        "pooling_strategy": "mean",
        "ndim": 384,
    },
    "bge-small": {
        "repo": "neuralmagic/bge-small-en-v1.5-quant",
        "path_in_repo": "model.onnx",
        "max_length": 512,
        "pooling_strategy": "first",
        "ndim": 384,
    },
    "multilingual-e5-small": {
        "repo": "intfloat/multilingual-e5-small",
        "path_in_repo": "onnx/model.onnx",
        "max_length": 512,
        "pooling_strategy": "mean",
        "ndim": 384,
    },
    # 12 layers, 768-dim
    "bge-base": {
        "repo": "neuralmagic/bge-base-en-v1.5-quant",
        "path_in_repo": "model.onnx",
        "max_length": 512,
        "pooling_strategy": "first",
        "ndim": 768,
    },
    # 24 layers, 1024-dim
    "bge-large": {
        "repo": "neuralmagic/bge-large-en-v1.5-quant",
        "path_in_repo": "model.onnx",
        "max_length": 512,
        "pooling_strategy": "first",
        "ndim": 1024,
    },
    "mixedbread-large": {
        "repo": 'mixedbread-ai/mxbai-embed-large-v1',
        "path_in_repo": "onnx/model_quantized.onnx",
        "max_length": 512,
        "pooling_strategy": "first",
        "normalize": True,
        "ndim": 1024
    },
    ### SPARSE MODELS ###
    "distilbert-splade": {
        "repo": "andersonbcdefg/distilbert-splade-onnx",
        "path_in_repo": "model.onnx",
        "max_length": 512,
        "pooling_strategy": "max"
    }
}
